#define NO_HDR_COMPONENTS

#include "image_data.h"

#include "image_io.h"

#ifndef NO_HDR_COMPONENTS
#include "camera_response.h"
#endif

#include <cstdlib>
#include <cstring>
#include <cmath>

#include <vector>
#include <string>
#include <iostream>
#include <algorithm>

const uint32_t WF_INCLUDE_LOW(0x1);
const uint32_t WF_INCLUDE_HIGH(0x2);
const uint32_t WF_UNIFORM_WEIGHT(0x4);

// Masks for all bits up to n.  Quick and dirty hack to get default "all" masks
// for different numbers of layers.
const uint32_t ImageMasks[32] = {
    0x00000001, 0x00000003, 0x00000007, 0x0000000f, 
    0x0000001f, 0x0000003f, 0x0000007f, 0x000000ff,
    0x000001ff, 0x000003ff, 0x000007ff, 0x00000fff,
    0x00001fff, 0x00003fff, 0x00007fff, 0x0000ffff,
    0x0001ffff, 0x0003ffff, 0x0007ffff, 0x000fffff, 
    0x001fffff, 0x003fffff, 0x007fffff, 0x00ffffff, 
    0x01ffffff, 0x03ffffff, 0x07ffffff, 0x0fffffff, 
    0x1fffffff, 0x3fffffff, 0x7fffffff, 0xffffffff};

// constants for pixel classifying.  Based on 8 bit input data.
const uint16_t PIXEL_SATURATION_THRESHOLD(64512);
const uint16_t PIXEL_DARK_THRESHOLD(1024);

////////////////////////////////////////////////////////////////////////////////
s3d_merge::ImageData::ImageData(const char *fn, s3d_merge::Image16 *image, double ev)
    : m_name(fn)
    , m_image(image)
    , m_ev(ev)
    , m_multiplier(pow (2, ev))
{
}

s3d_merge::ImageData::~ImageData()
{
    if(m_image)
    {
        delete m_image;
    }
}

double 
s3d_merge::ImageData::EV(void) const
{
    return m_ev;
}

double 
s3d_merge::ImageData::Multiplier(void) const
{
    return m_multiplier;
}

const 
s3d_merge::Image16 *
s3d_merge::ImageData::Image(void) const
{
    return m_image;
}

s3d_merge::ImageData::PixelCategory 
s3d_merge::ImageData::ClassifyValue(uint16_t pixel)
{
    PixelCategory result(PIXEL_INVALID);
    if(pixel >= PIXEL_SATURATION_THRESHOLD)
    {
        result = PIXEL_SATURATED;
    }
    if(pixel <= PIXEL_DARK_THRESHOLD)
    {
        result = PIXEL_DARK;
    }
    else
    {
        result = PIXEL_IN_RANGE;
    }

    return result;
}

s3d_merge::ImageData::PixelCategory 
s3d_merge::ImageData::ClassifyPixel(uint32_t x, uint32_t y, uint32_t channel) const
{
    PixelCategory result(PIXEL_INVALID);

    if(m_image &&  
       (x < m_image->get_width()) && 
       (y < m_image->get_height()) &&
       (channel < m_image->get_channels()))
    {
        uint16_t pixel = m_image->get_pixel(x, y)[channel];

        result = ClassifyValue(pixel);
    }

    return result;
}


double
s3d_merge::ImageData::GetWeight(uint32_t x, uint32_t y, 
                                uint32_t channel, uint32_t flags) const
{
    double result(0.0);


    if(flags & WF_UNIFORM_WEIGHT)
    {
        return 1.0;
    }
    else
    {
        double val = double(m_image->get_pixel(x, y)[channel]) / 65535.0;
        double dfc = val - 0.5; // distance from centre
        uint32_t chk_flag(0);
        if(val > 0.5)
        {
            chk_flag = WF_INCLUDE_HIGH;
        }
        else
        {
            chk_flag = WF_INCLUDE_LOW;
        }

        if(flags & chk_flag)
        {
            dfc*= 0.98;
        }
        else
        {
            dfc/= 0.98;
        }

        result = (1.0 - (dfc * dfc));
        if(result < 0.0) result = 0.0;
    }
    return result;
}
/*
double
s3d_merge::ImageData::GetValue(uint32_t x, uint32_t y, uint32_t channel,
                               const s3d_merge::CameraResponse &response) const
{
    double result(0);

    if(m_image &&  
       (x < m_image->get_width()) && 
       (y < m_image->get_height()) &&
       (channel < m_image->get_channels()))
    {
        result = response.Invert(channel, m_image->get_pixel(x, y)[channel]) *
                 m_multiplier;
    }

    return result;
}*/

double
s3d_merge::ImageData::GetDoubleValue(uint32_t x, uint32_t y, uint32_t channel) const
{
    double result(0.0);

    if(m_image &&  
       (x < m_image->get_width()) && 
       (y < m_image->get_height()) &&
       (channel < m_image->get_channels()))
    {
        result = m_image->get_pixel(x, y)[channel];
        result/= 65536;
    }
    return result;
}

uint16_t 
s3d_merge::ImageData::GetValue(uint32_t x, uint32_t y, uint32_t channel) const
{
    uint16_t result(0);
    if(m_image &&  
       (x < m_image->get_width()) && 
       (y < m_image->get_height()) &&
       (channel < m_image->get_channels()))
    {
        result = m_image->get_pixel(x, y)[channel];
    }
    return result;
}

////////////////////////////////////////////////////////////////////////////////
