#include "match.h"

#include <stdint.h>
#include <cmath>
#include <vector>
#include <algorithm>
#include "kdtree.h"
#include "constants.h"

double 
s3d_merge::CalculateAngleThreshold(double angle)
{
     double x, y, threshold;
     angle*=(s3d_math::pi / 180.0);
     y = sin(angle);
     x = 1.0 - cos(angle);
     threshold = sqrt((x*x)+(y*y));
     return threshold;
}

void
s3d_merge::BuildTriangles3D(const s3d_merge::KDTree3D &tree, 
                      const std::vector<s3d_merge::Point3D> &points,
                      double min_angle,
                      double max_angle,
                      size_t triangles_per_point,  // ignore for now
                      std::vector<Triangle> &triangles)
{
    // Build triangles.

    double min_threshold = CalculateAngleThreshold(min_angle);
    double max_threshold = CalculateAngleThreshold(max_angle);

    {
        std::vector<size_t> found;

        for(size_t i=0;i<points.size();i++)
        {
            found.clear();
            {
                std::vector<size_t> tmp;
                tree.FindItemsInRange(points[i], max_threshold, tmp);

                for(size_t j=0;j<tmp.size();j++)
                {
                    if(Point3D::Distance(points[i], points[tmp[j]]) >= min_threshold)
                    {
                        found.push_back(tmp[j]);
                    }
                }
            }

            std::sort(found.begin(), found.end());
    
            if(found.size() > 2)
            {
                triangles.push_back(Triangle(found[0], found[1], found[2]));
            }
        }
    }
}

// Brute force cubic evil!
void
s3d_merge::BuildTriangles2D(const s3d_merge::KDTree2D &tree, 
                      const std::vector<s3d_merge::Point2D> &points,
                      double min_radius,
                      double max_radius,
                      size_t triangles_per_point,  // ignore for now
                      std::vector<Triangle> &triangles)
{
    std::vector<size_t> found;

    for(size_t i=0;i<points.size();i++)
    {
        found.clear();
        {
            std::vector<size_t> tmp;
            tree.FindItemsInRange(points[i], max_radius, tmp);

            for(size_t j=0;j<tmp.size();j++)
            {
                if(Point2D::Distance(points[i], points[tmp[j]]) >= min_radius)
                {
                    found.push_back(tmp[j]);
                }
            }
        }

        std::sort(found.begin(), found.end());

        if(found.size() > 2)
        {
            triangles.push_back(Triangle(found[0], found[1], found[2]));
        }
    }
}

void
s3d_merge::MatchTriangleShapes(const Triangle &t3d, const std::vector<s3d_merge::Point3D> p3d,
                    const Triangle &t2d, const std::vector<s3d_merge::Point2D> p2d,
                    Triangle &out_3d, 
                    Triangle &out_2d)
{
    double lengths3d[3];
    double lengths2d[3];

    {
        s3d_merge::Point3D p1, p2, p3;
        p1 = p3d[t3d.m_points[0]];
        p1 = p3d[t3d.m_points[1]];
        p2 = p3d[t3d.m_points[2]];
        s3d_merge::Point3D centroid((p1 + p2 + p3) / 3.0);
        centroid.Normalise();

        s3d_merge::Point3D tp1(p1 - centroid * (s3d_merge::Point3D::DotProduct(p1, centroid)));
        s3d_merge::Point3D tp2(p2 - centroid * (s3d_merge::Point3D::DotProduct(p2, centroid)));
        s3d_merge::Point3D tp3(p3 - centroid * (s3d_merge::Point3D::DotProduct(p3, centroid)));

        lengths3d[0] = Point3D::Distance(tp1, tp2);
        lengths3d[1] = Point3D::Distance(tp2, tp3);
        lengths3d[2] = Point3D::Distance(tp1, tp3);
    }

    {
        s3d_merge::Point2D p1, p2, p3;

        p1 = p2d[t2d.m_points[0]];
        p1 = p2d[t2d.m_points[1]];
        p2 = p2d[t2d.m_points[2]];

        lengths2d[0] = Point2D::Distance(p1, p2);
        lengths2d[1] = Point2D::Distance(p2, p3);
        lengths2d[2] = Point2D::Distance(p1, p3);
    }

    //   Sort so that:
    //               
    // shortest = C  c-.   B = second longest, despite my ascii art
    //             /    \
    //            a------b
    //                A = longest
    // Identify A,B,C a,b,c


    size_t lidx3d[3], lidx2d[3];

    lidx2d[0] = lidx2d[1] = lidx2d[2] = 0;
    lidx3d[0] = lidx3d[1] = lidx3d[2] = 0;

    // 3D length indices.
    {
        float min, max;
        min = max = lengths3d[0];

        for(size_t i=1;i<3;i++)
        {
            if(lengths3d[i] <= min)
            {
                min = lengths3d[i];
                lidx3d[2] = i;
            }
            if(lengths3d[i] >= max)
            {
                max = lengths3d[i];
                lidx3d[0] = i;
            }
        }
        if(lidx3d[0] == 0)
        {
            if(lidx3d[2] == 1)
            {
                lidx3d[1] = 2;
            }
            else
            {
                lidx3d[1] = 1;
            }
        }
        else if(lidx3d[0] == 1)
        {
            if(lidx3d[2] == 2)
            {
                lidx3d[1] = 0;
            }
            else
            {
                lidx3d[1] = 2;
            }
        }
        else
        {
            if(lidx3d[2] == 1)
            {
                lidx3d[1] = 0;
            }
            else
            {
                lidx3d[1] = 1;
            }
        }
    }

    // 2D length indices.
    {
        float min, max;
        min = max = lengths2d[0];

        for(size_t i=1;i<3;i++)
        {
            if(lengths2d[i] <= min)
            {
                min = lengths2d[i];
                lidx2d[2] = i;
            }
            if(lengths2d[i] >= max)
            {
                max = lengths2d[i];
                lidx2d[0] = i;
            }
        }

        if(lidx2d[0] == 0)
        {
            if(lidx2d[2] == 1)
            {
                lidx2d[1] = 2;
            }
            else
            {
                lidx2d[1] = 1;
            }
        }
        else if(lidx2d[0] == 1)
        {
            if(lidx2d[2] == 2)
            {
                lidx2d[1] = 0;
            }
            else
            {
                lidx2d[1] = 2;
            }
        }
        else
        {
            if(lidx2d[2] == 1)
            {
                lidx2d[1] = 0;
            }
            else
            {
                lidx2d[1] = 1;
            }
        }
    }

    size_t pidx3d[3], pidx2d[3];

    // Identify which point is which.
    for(size_t i=0;i<3;i++)
    {
        if(((lidx3d[0] == i) && (lidx3d[2] == (i+2%3))) ||
           ((lidx3d[2] == i) && (lidx3d[0] == (i+2%3))))
        {
            pidx3d[i] = 0;
        }
        if(((lidx3d[0] == i) && (lidx3d[1] == (i+2%3))) ||
           ((lidx3d[1] == i) && (lidx3d[0] == (i+2%3))))
        {
            pidx3d[i] = 1;
        }
        else
        {
            pidx3d[i] = 2;
        }
    }

    // Identify which point is which.
    for(size_t i=0;i<3;i++)
    {
        if(((lidx2d[0] == i) && (lidx2d[2] == (i+2%3))) ||
           ((lidx2d[2] == i) && (lidx2d[0] == (i+2%3))))
        {
            pidx2d[i] = 0;
        }
        if(((lidx2d[0] == i) && (lidx2d[1] == (i+2%3))) ||
           ((lidx2d[1] == i) && (lidx2d[0] == (i+2%3))))
        {
            pidx2d[i] = 1;
        }
        else
        {
            pidx2d[i] = 2;
        }
    }

    // Now output.

    for(size_t i=0;i<3;i++)
    {
        out_3d.m_points[pidx3d[i]] = t3d.m_points[i];
        out_2d.m_points[pidx2d[i]] = t2d.m_points[i];
    }
    
} 


s3d_merge::Point2D
s3d_merge::GetTriangleShape3D(const s3d_merge::Point3D &p1, const s3d_merge::Point3D &p2, const s3d_merge::Point3D &p3)
{
    double lengths[3];

    s3d_merge::Point3D centroid((p1 + p2 + p3) / 3.0);
    centroid.Normalise();

    // Flatten points onto plane normal to centroid.

    s3d_merge::Point3D tp1(p1 - centroid * (s3d_merge::Point3D::DotProduct(p1, centroid)));
    s3d_merge::Point3D tp2(p2 - centroid * (s3d_merge::Point3D::DotProduct(p2, centroid)));
    s3d_merge::Point3D tp3(p3 - centroid * (s3d_merge::Point3D::DotProduct(p3, centroid)));

    lengths[0] = Point3D::Distance(tp1, tp2);
    lengths[1] = Point3D::Distance(tp2, tp3);
    lengths[2] = Point3D::Distance(tp1, tp3);

    // Dodgy sort.
    if (lengths[0] < lengths[1]) std::swap(lengths[0], lengths[1]);
    if (lengths[1] < lengths[2]) std::swap(lengths[1], lengths[2]);
    if (lengths[0] < lengths[1]) std::swap(lengths[0], lengths[1]);

    Point2D result;
    result[0] = lengths[1]/lengths[0];
    result[1] = lengths[2]/lengths[0];

    return result;
}

void
s3d_merge::BuildTriangleStats3D(const std::vector<Triangle> &triangles, 
                     const std::vector<s3d_merge::Point3D> &points,
                     std::vector<s3d_merge::Point2D> &shapes,
                     std::vector<s3d_merge::Point3D> &centroids)
{
    for(size_t i=0;i<triangles.size();i++)
    {
        // Triangle centroids
        {
            Point3D tmp_point;
            tmp_point = (points[triangles[i].m_points[0]] +
                         points[triangles[i].m_points[1]] +
                         points[triangles[i].m_points[2]]) / 3;
            tmp_point.Normalise();
            centroids.push_back(tmp_point);
            
        }
        shapes.push_back(GetTriangleShape3D(points[triangles[i].m_points[0]],
                                            points[triangles[i].m_points[1]],
                                            points[triangles[i].m_points[2]]));
    }
}

s3d_merge::Point2D
s3d_merge::GetTriangleShape2D(const s3d_merge::Point2D &p1, const s3d_merge::Point2D &p2, const s3d_merge::Point2D &p3)
{
    double lengths[3];
    lengths[0] = Point2D::Distance(p1, p2);
    lengths[1] = Point2D::Distance(p2, p3);
    lengths[2] = Point2D::Distance(p1, p3);

    // Dodgy sort.
    if (lengths[0] < lengths[1]) std::swap(lengths[0], lengths[1]);
    if (lengths[1] < lengths[2]) std::swap(lengths[1], lengths[2]);
    if (lengths[0] < lengths[1]) std::swap(lengths[0], lengths[1]);

    Point2D result;
    result[0] = lengths[1]/lengths[0];
    result[1] = lengths[2]/lengths[0];

    return result;
}

void
s3d_merge::BuildTriangleStats2D(const std::vector<Triangle> &triangles, 
                     const std::vector<s3d_merge::Point2D> &points,
                     std::vector<s3d_merge::Point2D> &shapes,
                     std::vector<s3d_merge::Point2D> &centroids)
{
    for(size_t i=0;i<triangles.size();i++)
    {
        // Triangle centroids
        {
            Point2D tmp_point;
            tmp_point = (points[triangles[i].m_points[0]] +
                         points[triangles[i].m_points[1]] +
                         points[triangles[i].m_points[2]]) / 3;
            tmp_point.Normalise();
            centroids.push_back(tmp_point);
            
        }

        shapes.push_back(GetTriangleShape2D(points[triangles[i].m_points[0]],
                                            points[triangles[i].m_points[1]],
                                            points[triangles[i].m_points[2]]));
    }
}

