#include "transform.h"
#include "constants.h"
#include <cmath>

////////////////////////////////////////////////////////////////////////////////
// Base class
s3d_merge::Transform::Transform()
    : m_data(0)
{
}

s3d_merge::Transform::~Transform()
{
    if(m_data)
    {
        delete[] m_data;
    }
}

void
s3d_merge::Transform::SetParameters(const double *p)
{
    if(p)
    {
        for(size_t i=0;i<GetNumParameters();i++)
        {
            m_data[i] = p[i];
        }
        OnSetParameters();
    }
}

void
s3d_merge::Transform::GetParameters(double *p) const
{
    if(p)
    {
        for(size_t i=0;i<GetNumParameters();i++)
        {
            p[i] = m_data[i];
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
// rotate & translate
s3d_merge::RotateAndTranslate::RotateAndTranslate()
{
    m_data = new double[RAT_MAX];
    SetDefaults();
    OnSetParameters();
}

s3d_merge::RotateAndTranslate::RotateAndTranslate(const s3d_merge::RotateAndTranslate &other)
{
    m_data = new double[RAT_MAX];
    other.GetParameters(m_data);
    OnSetParameters();
}

s3d_merge::RotateAndTranslate::~RotateAndTranslate()
{
}

const s3d_merge::RotateAndTranslate &
s3d_merge::RotateAndTranslate:: operator=(const s3d_merge::RotateAndTranslate &other)
{
    other.GetParameters(m_data);
    OnSetParameters();
    return *this;
}

void
s3d_merge::RotateAndTranslate::OnSetParameters()
{
    double angle = m_data[RAT_ANGLE];
    double tx    = m_data[RAT_TRANSLATE_X];
    double ty    = m_data[RAT_TRANSLATE_Y];

    double cos_a(cos(angle / s3d_math::one_eighty_over_pi));
    double sin_a(sin(angle / s3d_math::one_eighty_over_pi));

    m_fwd_i[0] =  cos_a;
    m_fwd_i[1] =  sin_a;

    m_inv_i[0] =  cos_a;
    m_inv_i[1] = -sin_a;

    m_fwd_j[0] = -sin_a;
    m_fwd_j[1] =  cos_a;

    m_inv_j[0] =  sin_a;
    m_inv_j[1] =  cos_a;

    m_t[0] = tx;
    m_t[1] = ty;
}

void
s3d_merge::RotateAndTranslate::SetDefaults()
{
    for(size_t i=0;i<RAT_MAX;i++)
    {
        m_data[i] = 0.0;
    }
    OnSetParameters();
}

size_t
s3d_merge::RotateAndTranslate::GetNumParameters() const
{
    return size_t(RAT_MAX);
}

void
s3d_merge::RotateAndTranslate::FwdTransformPoint(const s3d_merge::Point2D &in, s3d_merge::Point2D &out) const
{
    s3d_merge::Point2D tmp, tmp2;
    tmp  = in + m_t;                                  // fwd translate
    tmp2 = ((m_fwd_i * tmp[0]) + (m_fwd_j * tmp[1])); // fwd rotate
    out  = tmp2;
}

void
s3d_merge::RotateAndTranslate::InvTransformPoint(const s3d_merge::Point2D &in, s3d_merge::Point2D &out) const
{
    s3d_merge::Point2D tmp, tmp2;
    tmp  = ((m_inv_i * in[0]) + (m_inv_j * in[1])); // inv rotate
    tmp2 = tmp - m_t;                               // inv translate
    out  = tmp2;
}

////////////////////////////////////////////////////////////////////////////////
// Lens distort

s3d_merge::LensDistort::LensDistort()
{
    m_data = new double[LD_MAX];


    m_centre[0] = m_centre[1] = 0.5;
    SetDefaults();
}

s3d_merge::LensDistort::LensDistort(uint32_t x_res, uint32_t y_res)
{
    m_data = new double[LD_MAX];


    m_centre[0] = x_res;
    m_centre[1] = y_res;
    m_centre*=    0.5;

    SetDefaults();
}

/*s3d_merge::LensDistort::LensDistort(const LensDistort &other)
{
    m_data = new double[LD_MAX];
}*/

s3d_merge::LensDistort::~LensDistort()
{
}

/*
const s3d_merge::LensDistort &
s3d_merge::LensDistort::operator=(const LensDistort &other);
*/
void
s3d_merge::LensDistort::OnSetParameters()
{
    m_radius = sqrt(Point2D::DotProduct(m_centre, m_centre));

    double dx, dy;
    double lx, ly;
    
    //dx = fabs(m_data[LD_X]);
    //dy = fabs(m_data[LD_Y]); 
    dx = dy = 0.0;

    // Move offset.
    lx = m_centre[0] + dx;
    ly = m_centre[1] + dy;

    m_radius = sqrt((lx*lx)+(ly*ly));

    m_offset[0] = 0.0;
    m_offset[1] = 0.0;
    //m_offset[0] = m_data[LD_X];
    //m_offset[1] = m_data[LD_Y];
}

void
s3d_merge::LensDistort::SetDefaults()
{
    for(size_t i=0;i<LD_MAX;i++)
    {
        m_data[i] = 0.0;
    }
    OnSetParameters();
}

size_t
s3d_merge::LensDistort::GetNumParameters() const
{
    return size_t(LD_MAX);
}

void    
s3d_merge::LensDistort::FwdTransformPoint(const Point2D &in, Point2D &out) const
{
    Point2D tmp;
    tmp = (in - m_centre) + m_offset;
    tmp/= m_radius;

    double r = sqrt(Point2D::DotProduct(tmp, tmp));

    if(r > 1.0e-5)
    {
        tmp/= r;
        r = FwdRadialTerm(r);
        tmp*= r;

        out = ((tmp * m_radius) + m_centre) - m_offset;
    }
    else
    {
        out = in;
    }
}

void
s3d_merge::LensDistort::InvTransformPoint(const Point2D &in, Point2D &out) const
{
    Point2D tmp;
    tmp = (in - m_centre) + m_offset;
    tmp/= m_radius;

    double r = sqrt(Point2D::DotProduct(tmp, tmp));

    if(r > 1.0e-5) // good for 10k wide image.
    {
        tmp/= r;
        r = InvRadialTerm(r);
        tmp*= r;
        tmp-= m_offset;

        out = ((tmp * m_radius) + m_centre) - m_offset;
    }
    else
    {
        out = in;
    }
}

double
s3d_merge::LensDistort::FwdRadialTerm(double r) const
{
    double out;
    double r2 = r*r;
    double r3 = r*r2;
    double r4 = r2*r2;
    double r5 = r*r4;

    double k1(m_data[LD_K1]);
    double k2(m_data[LD_K2]);
    double k3(m_data[LD_K3]);
    double k4(m_data[LD_K4]);

    //printf("%g, %g\n", k1, k2);
    //k1 = k2 = 0;
    out = ((r + (k1*r2) + (k2*r3) + (k3 * r4) + (k4*r5)) / (1.0+k1+k2+k3+k4)); // Force to be between 0 and 1

    return out;
}

double
s3d_merge::LensDistort::InvRadialTerm(double r) const
{
    // Newtons method.

    double x(r);
    const double dx(1.0e-6);
    const double dx2(dx*2.0);

    double xpdx, xmdx;
    double fx, fxpdx, fxmdx;
    double dfdx;

    size_t i;
    for(i=0;i<50;i++)
    {
        xpdx = x+dx;
        xmdx = x-dx;
        fx = FwdRadialTerm(x) - r;

        if(fabs(fx) < 1.0e-6) break;

        fxpdx = FwdRadialTerm(xpdx) - r;
        fxmdx = FwdRadialTerm(xmdx) - r;

        dfdx = (fxpdx - fxmdx)/dx2;

        x-= fx/dfdx;
    }

    return x;
}
////////////////////////////////////////////////////////////////////////////////
s3d_merge::TransformAndLens::TransformAndLens(uint32_t x_res, uint32_t y_res)
    : m_lens(x_res, y_res)
{
    m_data = new double[GetNumParameters()];
}

s3d_merge::TransformAndLens::~TransformAndLens()
{
}

void
s3d_merge::TransformAndLens::GetParameters(double *p) const
{
    if(p)
    {
        size_t i=0;
        m_lens.GetParameters(&p[i]);
        i+= m_lens.GetNumParameters();

        m_move.GetParameters(&p[i]);
        i+= m_move.GetNumParameters();
    }
}

void    
s3d_merge::TransformAndLens::OnSetParameters()
{
    size_t i=0;

    m_lens.SetParameters(&m_data[i]);
    i+= m_lens.GetNumParameters();

    m_move.SetParameters(&m_data[i]);
    i+= m_move.GetNumParameters();
}

void
s3d_merge::TransformAndLens::SetDefaults()
{
    for(size_t i=0;i<GetNumParameters();i++)
    {
        m_data[i] = 0.0;
    }
    OnSetParameters();
}

size_t
s3d_merge::TransformAndLens::GetNumParameters() const
{
    return m_lens.GetNumParameters() + m_move.GetNumParameters();
}

void    
s3d_merge::TransformAndLens::FwdTransformPoint(const Point2D &in, Point2D &out) const
{
    Point2D tmp1, tmp2;

    m_lens.FwdTransformPoint(in, tmp1);   // Undo distortion of this lens.
    m_move.FwdTransformPoint(tmp1, tmp2); // xform.
    m_lens.InvTransformPoint(tmp2, out);  // Distortion of other lens.
}

void
s3d_merge::TransformAndLens::InvTransformPoint(const Point2D &in, Point2D &out) const
{
    Point2D tmp1, tmp2;

    m_lens.FwdTransformPoint(in, tmp1);   // Undo distortion of other lens.
    m_move.InvTransformPoint(tmp1, tmp2); // xform.
    m_lens.InvTransformPoint(tmp2, out);  // Distortion of this lens.
}

